<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">
  <xsl:output method="html" doctype-public="-//W3C//DTD HTML 4.01 Transitional//EN" encoding="UTF-8" doctype-system="http://www.w3.org/TR/html4/loose.dtd"/>
  <xsl:strip-space elements="*"/>

  <!-- TO ADAPT / INTEGRATE
        - handling of assumptions
        - better output for split / splitProc
    -->
  <xsl:template name="ltsStyles">
    <style type="text/css">
      .prevar { color: green; font-style: italic; }
      .postvar { color: brown; font-style: italic; }
      .error { color: red; }
      .node { color: blue; }
      .sharpnode { color: red; }
      .sharpnode:after { font-size: small; vertical-align: super; content: "#"; }
      .transition:before { font-size: medium; vertical-align: 15%; color: black; content: "&#8594;"; }
      .sharptransition:before { font-size: medium; vertical-align: 15%; color: black; content: "&#8594;"; }
      .transition { font-size: small; vertical-align: -15%; font-style: italic; }
      .sharptransition { font-size: small; vertical-align: -15%; font-style: italic; color: red; }
      .sharptransition:after { font-size: x-small; content: "#"; vertical-align: super; }
    </style>
  </xsl:template>

  <xsl:template match="/certificationProblem">
    <xsl:variable name="mode">
      <xsl:choose>
        <xsl:when test="./@type='safety'">
          Safety
        </xsl:when>
        <xsl:otherwise>
          Termination
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <html>
      <head>
        <title>
          <xsl:copy-of select="$mode"/> Proof
        </title>
        <xsl:call-template name="ltsStyles"/>
      </head>
      <body>
        <h1>
          <xsl:copy-of select="$mode"/> Problem
        </h1>
        <h2>Input</h2>
        <xsl:apply-templates select="input"/>
        <xsl:for-each select="proof">
          <xsl:for-each select="ltsSafetyProof/*">
            <h2>Safey Proof</h2>
            <xsl:call-template name="ltsSafetyProof">
              <xsl:with-param name="indent" select="position()"/>
            </xsl:call-template>
          </xsl:for-each>
          <xsl:for-each select="ltsTerminationProof/*">
            <h2>Termination Proof</h2>
            <xsl:call-template name="ltsTerminationProof">
              <xsl:with-param name="indent" select="position()"/>
            </xsl:call-template>
          </xsl:for-each>
          <xsl:if test="not(ltsSafetyProof or ltsTerminationProof)"><!-- to be removed -->
            <xsl:for-each select="*[not(self::input or self::error)]">
              <xsl:call-template name="ltsTerminationProof">
                <xsl:with-param name="indent" select="position()"/>
              </xsl:call-template>
            </xsl:for-each>
          </xsl:if>
        </xsl:for-each>
      </body>
    </html>
  </xsl:template>

  <xsl:template match="error">
    <xsl:apply-templates/>
  </xsl:template>

  <xsl:template match="locationId">
    <span class="node">
      <xsl:value-of select="text()"/>
    </span>
  </xsl:template>
  <xsl:template name="locationDuplicate">
    <span class="sharpnode">
      <xsl:value-of select="text()"/>
    </span>
  </xsl:template>
  <xsl:template match="locationDuplicate">
    <xsl:call-template name="locationDuplicate"/>
  </xsl:template>

  <xsl:template match="ltsSafetyInput">
    <xsl:apply-templates select="lts"/>
      Error locations
      <ul>
        <xsl:for-each select="error">
          <li><xsl:apply-templates/></li>
        </xsl:for-each>
      </ul>
  </xsl:template>
  
  <xsl:template match="lts">
    Integer Transition System
    <ul>
      <li>
        Initial Location: <xsl:apply-templates select="initial"/>
      </li>
      <li>
        Transitions: (<span class="prevar">pre-variables</span> and <span class="postvar">post-variables</span>)
        <table>
          <xsl:for-each select="transition">
            <tr style="vertical-align:baseline">
              <td align="right">
                <xsl:apply-templates select="source"/>
              </td>
              <td align="left">
                <xsl:apply-templates select="transitionId"/>
              </td>
              <td align="left">
                <xsl:apply-templates select="target"/>:
              </td>
              <td> </td>
              <td align="left">
                <xsl:choose>
                  <xsl:when test="formula">
                    <xsl:apply-templates select="formula"/>
                  </xsl:when>
                  <xsl:otherwise>
                    skip
                  </xsl:otherwise>
                </xsl:choose>
              </td>
            </tr>
          </xsl:for-each>
        </table>
      </li>
    </ul>
  </xsl:template>
  
  <xsl:template match="transition">
    <span style="vertical-align: baseline">
      <xsl:apply-templates select="source"/>
      <xsl:text> </xsl:text>
      <xsl:apply-templates select="*[1]"/>
      <xsl:text> </xsl:text>
      <xsl:apply-templates select="target"/>:
      <xsl:choose>
        <xsl:when test="formula">
          <xsl:apply-templates select="formula"/>
        </xsl:when>
        <xsl:otherwise>
          skip
        </xsl:otherwise>
      </xsl:choose>
    </span>
  </xsl:template>

  <xsl:template match="conjunction">
    <xsl:if test="count(*) = 0">
      TRUE
    </xsl:if>
    <xsl:for-each select="*">
      <xsl:if test="position() > 1">
        <xsl:text> &#8743; </xsl:text>
      </xsl:if>
      <xsl:apply-templates select="."></xsl:apply-templates>
    </xsl:for-each>
  </xsl:template>

  <xsl:template match="disjunction">
    <xsl:if test="count(*) = 0">
      FALSE
    </xsl:if>
    <xsl:for-each select="*">
      <xsl:if test="position() > 1">
        <xsl:text> &#8744; </xsl:text>
      </xsl:if>
      <xsl:apply-templates select="."></xsl:apply-templates>
    </xsl:for-each>
  </xsl:template>

  <xsl:template match="leq">
    <xsl:apply-templates select="*[1]"></xsl:apply-templates>
    <xsl:text> </xsl:text>
    &#8804;
    <xsl:text> </xsl:text>
    <xsl:variable name="rhs" select="*[2]"/>
    <xsl:choose>
      <xsl:when test="$rhs">
        <xsl:apply-templates select="$rhs"></xsl:apply-templates>
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>0</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="less">
    <xsl:apply-templates select="*[1]"></xsl:apply-templates>
    <xsl:text> &lt; </xsl:text>
    <xsl:apply-templates select="*[2]"></xsl:apply-templates>
  </xsl:template>
  
  <xsl:template match="eq">
    <xsl:apply-templates select="*[1]"></xsl:apply-templates>
    <xsl:text> = </xsl:text>
    <xsl:apply-templates select="*[2]"></xsl:apply-templates>
  </xsl:template>
  
  <xsl:template match="constant">
    <xsl:param name="sgn" select="1"/>
    <xsl:param name="one" select="1"/>
    <xsl:if test="$sgn and text() &lt; 0">&#8722;</xsl:if>
    <xsl:variable name="ac" select="format-number(text(),'0;0')"></xsl:variable>
    <xsl:if test="$one or $ac!='1'">
      <xsl:value-of select="$ac"/>
    </xsl:if>
  </xsl:template>

  <xsl:template match="post">
    <span class="postvar">
      <xsl:value-of select="variableId/text()"/>
    </span>
  </xsl:template>
  <xsl:template match="variableId">
    <xsl:choose>
      <xsl:when test="post">
      </xsl:when>
      <xsl:otherwise>
        <span class="prevar">
          <xsl:value-of select="text()"/>
        </span>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="product">
    <xsl:param name="sgn" select="1"/>
    <xsl:choose>
      <xsl:when test="name(*[1]) = 'constant' and (name(*[2]) = 'post' or name(*[2]) = 'variableId')">
        <xsl:apply-templates select="*[1]">
          <xsl:with-param name="sgn" select="$sgn"/>
          <xsl:with-param name="one" select="0"/>
        </xsl:apply-templates>
        <xsl:apply-templates select="*[2]"/>
        <xsl:for-each select="*[position() > 2]">
          <xsl:text>&#8901;</xsl:text>
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </xsl:when>
      <xsl:otherwise>
        <xsl:for-each select="*">
          <xsl:if test="position() != 1">&#8901;</xsl:if>
          <xsl:apply-templates select=".">
            <xsl:with-param name="sgn" select="$sgn or position != 1"/>
          </xsl:apply-templates>
        </xsl:for-each>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="sum">
    <xsl:if test="not(node())">
      0
    </xsl:if>
    <xsl:for-each select="*">
      <xsl:choose>
        <xsl:when test="position() = 1">
          <xsl:apply-templates select=".">
            <xsl:with-param name="sgn" select="1"/>
          </xsl:apply-templates>
        </xsl:when>
        <xsl:when test="self::constant and ./text() &lt; 0">
          &#8722;
          <xsl:apply-templates select=".">
            <xsl:with-param name="sgn" select="0"/>
          </xsl:apply-templates>
        </xsl:when>
        <xsl:when test="self::product and local-name(*[1])='constant' and constant/text() &lt; 0">
          &#8722;
          <xsl:apply-templates select=".">
            <xsl:with-param name="sgn" select="0"/>
          </xsl:apply-templates>
        </xsl:when>
        <xsl:otherwise>
          + <xsl:apply-templates select="."/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>
  </xsl:template>

  <xsl:template match="linearImplicationHint">
    <xsl:choose>
      <xsl:when test="linearCombination">
        [<xsl:for-each select="linearCombination/*">
          <xsl:if test="position() > 1">, </xsl:if>
          <xsl:apply-templates select="."/>
        </xsl:for-each>]        
      </xsl:when>
      <xsl:when test="simplex">simplex</xsl:when>
    </xsl:choose>    
  </xsl:template>

  <xsl:template match="impact">
    <xsl:param name="indent"/>
    <h3>
      <xsl:value-of select="$indent"/> IMPACT Invariant Proof
    </h3>
    <ul>
      <li>
        initial node: <xsl:apply-templates select="*[1]"/>
      </li>
      <li>invariants:
        <table>
          <tr>
            <th>node</th>
            <th>location</th>
            <th style="text-align: left;">invariant</th>
          </tr>
          <xsl:for-each select="nodes/node">
            <tr style="vertical-align: baseline;">
              <td align="center">
                <xsl:apply-templates select="*[1]"/>
              </td>
              <td align="center">
                <xsl:apply-templates select="location"/>
              </td>
              <td align="left">
                <xsl:apply-templates select="invariant"/>
              </td>
            </tr>
          </xsl:for-each>
        </table>
      </li>
      <li>
        cover edges:
        <table>
          <xsl:for-each select="nodes/node/coverEdge">
            <tr>
              <td align="right">
                <xsl:apply-templates select="../*[1]"/>
              </td>
              <td align="left">
                <xsl:text>&#8594; </xsl:text>
                <xsl:apply-templates select="nodeId"/>
              </td>
              <td>
                <xsl:apply-templates select="*[2]"/>
              </td>
            </tr>
          </xsl:for-each>
        </table>
      </li>
      <li>
        transition edges:
        <table>
          <xsl:for-each select="nodes/node/children/child">
            <tr>
              <td align="right">
                <xsl:apply-templates select="../../*[1]"/>
              </td>
              <td align="left">
                <xsl:apply-templates select="*[1]"/>
                <xsl:text> </xsl:text>
                <xsl:apply-templates select="nodeId"></xsl:apply-templates>
              </td>
              <td>
                <xsl:if test="hints">
                  Hint:
                  <xsl:for-each select="hints/*">
                    <xsl:text> </xsl:text>
                    <xsl:apply-templates select="."/>
                  </xsl:for-each>
                </xsl:if>
              </td>
            </tr>
          </xsl:for-each>
        </table>
      </li>
    </ul>
  </xsl:template>

  <xsl:template match="transitionId">
    <span class="transition"><xsl:apply-templates/></span>
  </xsl:template>

  <xsl:template match="transitionDuplicate">
    <span class="sharptransition"><xsl:apply-templates/></span>
  </xsl:template>

  <xsl:template name="ltsTerminationProof">
    <xsl:param name="indent"/>
    <xsl:choose>
      <xsl:when test="self::switchToCooperationTermination">
        <h3>
          <xsl:value-of select="$indent"/> Switch to Cooperation Termination Proof
        </h3>
          We consider the following cutpoint-transitions:
          <table>
            <xsl:for-each select="cutPoints/cutPoint">
              <tr style="vertical-align: baseline">
                <td>
                  <xsl:apply-templates select="locationId"/>
                </td>
                <td>
                  <span class="transition">
                    <xsl:apply-templates select="skipId"/>
                  </span>
                  <xsl:text> </xsl:text>
                </td>
                <td>
                  <xsl:for-each select="locationId">
                    <xsl:call-template name="locationDuplicate"/>
                  </xsl:for-each>
                  <xsl:text>: </xsl:text>
                </td>
                <td>
                  <xsl:choose>
                    <xsl:when test="skipFormula">
                      <xsl:apply-templates select="skipFormula"/>
                    </xsl:when>
                    <xsl:otherwise>
                      SKIP
                    </xsl:otherwise>
                  </xsl:choose>
                </td>
              </tr>
            </xsl:for-each>
          </table>
          and for every transition <span class="transition">t</span>,
          a duplicate <span class="sharptransition">t</span> is considered.
        <xsl:for-each select="*[2]">
            <xsl:call-template name="cooperationProof">
              <xsl:with-param name="indent" select="concat($indent,'.',position())"/>
            </xsl:call-template>
        </xsl:for-each>
      </xsl:when>
      <xsl:when test="self::newInvariants">
        <h3>
          <xsl:value-of select="$indent"/> Invariant Updates
        </h3>
        <p>The following invariants are asserted.</p>
        <xsl:apply-templates select="invariants"/>
        <p>The invariants are proved as follows.</p>
        <xsl:apply-templates select="impact">
          <xsl:with-param name="indent" select="''"/>
        </xsl:apply-templates>
        <p></p>
        <xsl:for-each select="*[3]">
          <xsl:call-template name="ltsTerminationProof">
            <xsl:with-param name="indent" select="concat($indent,'.1')"/>
          </xsl:call-template>
        </xsl:for-each>
      </xsl:when>
      <xsl:otherwise>
        <h3>
          <xsl:value-of select="$indent"/> Unknown Termination Proof "<xsl:value-of select="name(.)"/>"!
        </h3>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="ltsTerminationProof">
    <xsl:for-each select="*">
      <xsl:call-template name="ltsTerminationProof">
        <xsl:with-param name="indent" select="position()"/>
      </xsl:call-template>
    </xsl:for-each>
  </xsl:template>

  <xsl:template name="ltsSafetyProof">
    <xsl:param name="indent"/>
    <xsl:choose>
      <xsl:when test="self::safetyViaInvariants">
        <h3>
          <xsl:value-of select="$indent"/> Invariant Updates
        </h3>
        <p>The following invariants show that the error locations are unreachable.</p>
        <xsl:apply-templates select="invariants"/>
        <p>The invariants are proved as follows.</p>
        <xsl:apply-templates select="impact">
          <xsl:with-param name="indent" select="''"/>
        </xsl:apply-templates>        
      </xsl:when>
      <xsl:otherwise>
        <h3>
          <xsl:value-of select="$indent"/> Unknown Safety Proof "<xsl:value-of select="name(.)"/>"!
        </h3>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="ltsSafetyProof">
    <xsl:for-each select="*">
      <xsl:call-template name="ltsSafetyProof">
        <xsl:with-param name="indent" select="position()"/>
      </xsl:call-template>
    </xsl:for-each>
  </xsl:template>

  <xsl:template name="cooperationProof">
    <xsl:param name="indent"/>
    <xsl:choose>
      <xsl:when test="self::trivial">
        <h3>
          <xsl:value-of select="$indent"/> Trivial Cooperation Program
        </h3>
        <p>
          There are no more "sharp" transitions in the cooperation program.
          Hence the cooperation termination is proved.
        </p>
      </xsl:when>
      <xsl:when test="self::transitionRemoval">
        <h3>
          <xsl:value-of select="$indent"/> Transition Removal
        </h3>
        <p>
          The following (lexicographic) ranking functions are used, which are bounded by
          <xsl:text>[</xsl:text>
          <xsl:for-each select="bound/*">
            <xsl:if test="position() > 1">
              <xsl:text>, </xsl:text>
            </xsl:if>
          <xsl:apply-templates select="."/>
          </xsl:for-each>].
        </p>
        <table>
          <xsl:for-each select="rankingFunctions/rankingFunction">
            <tr>
              <td>
                <xsl:apply-templates select="location"/>:
              </td>
              <td>
                <xsl:for-each select="expression">
                  <xsl:text>[</xsl:text>
                  <xsl:for-each select="*">
                    <xsl:if test="position() > 1">
                      <xsl:text>, </xsl:text>
                    </xsl:if>
                    <xsl:apply-templates select="."/>
                  </xsl:for-each>
                  <xsl:text>]</xsl:text>
                </xsl:for-each>
              </td>
            </tr>
          </xsl:for-each>
        </table>
        <p>
          Removed transitions are:
          <xsl:for-each select="remove/*">
            <xsl:if test="position() > 1">
              <xsl:text>, </xsl:text>
            </xsl:if>
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </p>
        <xsl:if test="hints/*">
        <table>
          <tr>
            <th>Transition</th>
            <th>Bound hints</th>
            <th>Decrease hints</th>
          </tr>
          <xsl:for-each select="hints/*">
            <tr>
              <td>
                <xsl:apply-templates select="*[1]"/>
              </td>
              <td>
                <table>
                  <xsl:for-each select="boundHints/*">
                    <tr><td><xsl:apply-templates select="."/></td></tr>
                  </xsl:for-each>
                </table>
              </td>
              <td>
                <table>
                  <xsl:for-each select="decreaseHints/*">
                    <tr><td><xsl:apply-templates select="."/></td></tr>
                  </xsl:for-each>
                </table>
              </td>
            </tr>
          </xsl:for-each>
        </table>
        </xsl:if>
        <xsl:variable name="base">
          <xsl:choose>
            <xsl:when test="hints">4</xsl:when>
            <xsl:otherwise>3</xsl:otherwise>
          </xsl:choose>
        </xsl:variable>
        <xsl:for-each select="*">
          <xsl:if test="position() > $base">
            <xsl:call-template name="cooperationProof">
              <xsl:with-param name="indent" select="concat($indent,'.',position()-$base)"/>
            </xsl:call-template>
          </xsl:if>
        </xsl:for-each>
      </xsl:when>
      <xsl:when test="self::newInvariants">
        <h3>
          <xsl:value-of select="$indent"/> Invariant Updates
        </h3>
        <p>The following invariants are asserted.</p>
        <xsl:apply-templates select="invariants"/>
        <p>The invariants are proved as follows.</p>
        <xsl:apply-templates select="impact">
          <xsl:with-param name="indent" select="''"/>
        </xsl:apply-templates>
        <p></p>
        <xsl:for-each select="*[3]">
            <xsl:call-template name="cooperationProof">
              <xsl:with-param name="indent" select="concat($indent,'.1')"/>
            </xsl:call-template>
        </xsl:for-each>
      </xsl:when>
      <xsl:when test="self::locationAddition">
        <h3>
          <xsl:value-of select="$indent"/> Location Addition
        </h3>
        <p>
          The following skip-transition is inserted and corresponding redirections w.r.t. the old location are performed. 
        </p>
        <p>
          <xsl:apply-templates select="transition"/>
        </p>
        <xsl:for-each select="*[2]">
          <xsl:call-template name="cooperationProof">
            <xsl:with-param name="indent" select="concat($indent,'.1')"/>
          </xsl:call-template>
        </xsl:for-each>
      </xsl:when>
      <xsl:when test="self::freshVariableAddition">
        <h3>
          <xsl:value-of select="$indent"/> Fresh Variable Addition
        </h3>
        <p>
        The new variable <xsl:apply-templates select="*[1]"/> is introduced.
        The transition formulas are extended as follows:
        <table>
          <xsl:for-each select="additionalFormulas/additionalFormula">
            <tr><td><xsl:apply-templates select="*[1]"/>: </td><td><xsl:apply-templates select="*[2]"/></td></tr>
          </xsl:for-each>
        </table>
        </p>
          <xsl:for-each select="*[4]">
            <xsl:call-template name="cooperationProof">
              <xsl:with-param name="indent" select="concat($indent,'.1')"/>
            </xsl:call-template>
          </xsl:for-each>
        
      </xsl:when>
      <xsl:when test="self::sccDecomposition">
        <h3>
          <xsl:value-of select="$indent"/> SCC decomposition 
        </h3>
        <p>
          We consider subproblems for each of the <xsl:value-of select="count(sccWithProof)"/> SCC(s) of the call-graph.
          We number the SCCs as follows.
          <ol>
            <xsl:for-each select="sccWithProof">
              <li> SCC
                <xsl:text>{ </xsl:text>
                <xsl:for-each select="scc/*">
                  <xsl:if test="position() != 1">, </xsl:if>
                  <xsl:apply-templates select="."/>
                </xsl:for-each>
                <xsl:text> }</xsl:text>
              </li>
            </xsl:for-each>
          </ol>
          <xsl:for-each select="sccWithProof">
            <xsl:variable name="num" select="position()"/>
            <xsl:for-each select="*[2]">
              <xsl:call-template name="cooperationProof">
                <xsl:with-param name="indent" select="concat($indent,'.',$num)"/>
              </xsl:call-template>
            </xsl:for-each>
          </xsl:for-each>          
        </p>        
      </xsl:when>
      <xsl:when test="self::cutTransitionSplit">
        <h3>
          <xsl:value-of select="$indent"/> Splitting of Cut-Transitions 
        </h3>
        <p>
          We consider subproblems for each of the following <xsl:value-of select="count(cutTransitionsWithProof)"/> sets of cut transitions.
          We number the SCCs as follows.
          <ol>
            <xsl:for-each select="cutTransitionsWithProof">
              <li> set of transitions:
                <xsl:text>{ </xsl:text>
                <xsl:for-each select="cutTransitions/*">
                  <xsl:if test="position() != 1">, </xsl:if>
                  <xsl:apply-templates select="."/>
                </xsl:for-each>
                <xsl:text> }</xsl:text>
              </li>
            </xsl:for-each>
          </ol>
          <xsl:for-each select="cutTransitionsWithProof">
            <xsl:variable name="num" select="position()"/>
            <xsl:for-each select="*[2]">
              <xsl:call-template name="cooperationProof">
                <xsl:with-param name="indent" select="concat($indent,'.',$num)"/>
              </xsl:call-template>
            </xsl:for-each>
          </xsl:for-each>          
        </p>        
      </xsl:when>
      <xsl:otherwise>
        <h3>
          <xsl:value-of select="$indent"/> Unknown Cooperation Proof "<xsl:value-of select="name(.)"/>"!
        </h3>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="invariants">
    <table>
      <xsl:for-each select="invariant">
        <tr>
          <td>
            <xsl:apply-templates select="location"/>:
          </td>
          <td>
            <xsl:apply-templates select="formula"/>
          </td>
        </tr>
      </xsl:for-each>
    </table>
  </xsl:template>

</xsl:stylesheet>