theory LIA_Instance
  imports
  LIA_Checker
  LTS_Termination_Prover
  LTS_Safety_Prover
begin

no_notation Simplex.curr_val_satisfies_state ("\<Turnstile>")

context LIA begin

fun constant_exp where
  "constant_exp (Boolean True) = Fun Less [lia_const 0, lia_const 1]"
| "constant_exp (Boolean False) = Fun Less [lia_const 1, lia_const 0]"
| "constant_exp (Integer n) = lia_const n"

sublocale algebra_constant "interpret" "constant_exp"
proof
  show "vars_term (constant_exp d) = {}" for d by (cases d rule: constant_exp.cases, auto)
  show "\<lbrakk>constant_exp d\<rbrakk> \<alpha> = d" for d \<alpha> by (cases d rule: constant_exp.cases, auto)
qed

fun is_constant where
  "is_constant (Fun _ []) = True"
| "is_constant _ = False"

lemma is_constant:
  assumes "is_constant e" shows "\<lbrakk>e\<rbrakk>\<alpha> = \<lbrakk>e\<rbrakk>\<beta>"
proof(cases e)
  case (Fun f es) with assms show ?thesis by (cases es, auto)
qed (insert assms, auto)

sublocale art_checker \<Lambda> check_lia_clause check_lia_clause show_LIA_exp show_LIA_exp ..

sublocale pre_definability_checker \<Lambda> def_checker .

sublocale transition_definability_checker: transition_definability_checker \<Lambda> def_checker show_LIA_exp show_LIA_exp .. 

abbreviation "less_eq_formula s t \<equiv> Atom (Fun Lesseq [s,t])"
abbreviation "less_formula s t \<equiv> Atom (Fun Less [s,t])"

lemma less_eq_meaning:
  assumes \<alpha>: "assignment \<alpha>"
      and s: "s :\<^sub>f IntegerT"
      and t: "t :\<^sub>f IntegerT"
  shows "\<alpha> \<Turnstile> less_eq_formula s t \<longleftrightarrow> \<lbrakk>s\<rbrakk> \<alpha> \<le> \<lbrakk>t\<rbrakk> \<alpha>"
  using eval_types[OF s \<alpha>] eval_types[OF t \<alpha>] s t
  by (cases "\<lbrakk>s\<rbrakk>\<alpha>"; cases "\<lbrakk>t\<rbrakk>\<alpha>", auto simp: is_bool_def eval_true_def)

lemma less_meaning:
  assumes \<alpha>: "assignment \<alpha>"
      and s: "s :\<^sub>f IntegerT"
      and t: "t :\<^sub>f IntegerT"
  shows "\<alpha> \<Turnstile> less_formula s t \<longleftrightarrow> \<lbrakk>s\<rbrakk> \<alpha> < \<lbrakk>t\<rbrakk> \<alpha>"
  using eval_types[OF s \<alpha>] eval_types[OF t \<alpha>] s t
  by (cases "\<lbrakk>s\<rbrakk>\<alpha>"; cases "\<lbrakk>t\<rbrakk>\<alpha>", auto simp: is_bool_def eval_true_def)

interpretation transition_removal: transition_removal_bounded \<Lambda> check_lia_clause show_LIA_exp
  "op \<le> :: value \<Rightarrow> _ \<Rightarrow> _" "op <"
  IntegerT less_eq_formula less_formula is_constant
  by (unfold_locales; (fact less_eq_meaning less_meaning is_constant | simp add: All_less_Suc))


sublocale termination_checker: termination_checker \<Lambda> check_lia_clause check_lia_clause show_LIA_exp show_LIA_exp def_checker
  IntegerT "op \<le>" "op <" less_eq_formula less_formula is_constant
    by (unfold_locales)

definition check_termination where "check_termination \<equiv> termination_checker.check"

corollary check_termination: "isOK (check_termination Pi prf) \<Longrightarrow> lts_termination (lts_of Pi)"
  by (unfold check_termination_def, fact termination_checker.sound)

thm check_safety

end

thm LIA_logic.check_termination LIA_logic.check_safety

declare LIA_logic.constant_exp.simps[code]
  LIA_logic.check_termination_def[code]
  LIA_logic.check_safety_def[code]

fun change_transition_formula where "change_transition_formula f (Transition l r \<phi>) = (Transition l r (f \<phi>))" 

fun of_LIA'_lts :: "(LIA.sig',_,_,_,_) lts_impl \<Rightarrow> (LIA.sig,_,_,_,_) lts_impl" where
  "of_LIA'_lts (Lts_Impl I t lc) = (Lts_Impl I 
    (map (\<lambda> (i,tr). (i, change_transition_formula LIA.of_LIA_formula' tr)) t)  
    (map (\<lambda> (l,f). (l, LIA.of_LIA_formula' f)) lc))" 

definition lts_of' where "lts_of' = lts_of o of_LIA'_lts"

end