theory Lexicographic_Theory
imports
  Formula
  "$AFP/Show/Show_Instances"
  "$AFP/Deriving/Comparator_Generator/Compare_Generator"
  "$AFP/Certification_Monads/Check_Monad"
  "../Orderings/Non_Inf_Order"
  Show
begin

subsection {* Bounded lexicographic order *}

context ord_syntax begin

fun lex_less_bounded' where
  "lex_less_bounded' (b#bs) (l#ls) (r#rs) = ((l \<sqsubset> r \<or> (l \<sqsubseteq> r \<and> lex_less_bounded' bs ls rs)) \<and> b \<sqsubset> r)" |
  "lex_less_bounded' _ _ _ = False"

fun lex_less_eq where
  "lex_less_eq (l#ls) (r#rs) = (l \<sqsubset> r \<or> (l \<sqsubseteq> r \<and> lex_less_eq ls rs))" |
  "lex_less_eq Nil Nil   = True" |
  "lex_less_eq (_#_) Nil = False" |
  "lex_less_eq Nil (_#_) = False"

lemma lex_less_eqI1:
  assumes "length ls = length rs"
      and "\<And> i. i < length ls \<Longrightarrow> ls ! i \<sqsubseteq> rs ! i"
  shows "lex_less_eq ls rs"
  using assms proof(induct ls rs rule:lex_less_eq.induct)
  case (1 l ls r rs)
  { fix i assume "i < length ls" hence "Suc i < length (l # ls)" by auto
    from 1(3)[OF this] have "ls ! i \<sqsubseteq> rs ! i" by auto }
  hence "lex_less_eq ls rs" using 1(1,2) by auto (* the next step "thus" gets this, but we state the induction step explicitly *)
  thus ?case using 1(3)[of 0] by auto
qed auto

lemma lex_less_eqI2:
  assumes eqs:"l1 = h1 @ (e1#t1)" "l2 = h2 @ (e2#t2)" "length h1 = length h2"
      and all:"\<And> i. i < length h1 \<Longrightarrow> h1 ! i \<sqsubseteq> h2 ! i"
      and e:"e1 \<sqsubset> e2"
  shows "lex_less_eq l1 l2"
  using eqs all proof(induct h1 h2 arbitrary:l1 l2 rule:lex_less_eq.induct)
  case (1 l ls r rs)
  { fix i assume "i < length ls" hence "Suc i < length (l # ls)" by auto
    from 1(5)[OF this] have "less_eq (ls ! i) (rs ! i)" by auto }
  hence "lex_less_eq (ls @ e1 # t1) (rs @ e2 # t2)" using 1 by auto
  then show ?case using 1(5)[of 0] 1(2,3) by auto
qed (insert e, auto)

lemma lex_less_bounded'I [intro]:
  assumes eqs:"l1 = h1 @ (e1#t1)" "l2 = h2 @ (e2#t2)" "l3 = h3 @ (e3#t3)"
              "length h1 = length h2" "length h2 = length h3"
      and all:"\<And> i. i<length h1 \<Longrightarrow> h1 ! i \<sqsubseteq> h2 ! i \<and> h3 ! i \<sqsubset> h2 ! i"
      and e:" e1 \<sqsubset> e2"
      and bnd:"e3 \<sqsubset> e2"
  shows "lex_less_bounded' l3 l1 l2"
  using eqs all proof(induct h1 h2 h3 arbitrary:l1 l2 l3 rule:lex_less_bounded'.induct)
  case (1 l ls r rs b bs)
  { fix i assume "i < length ls" hence "Suc i < length (l # ls)" by auto
    from 1(7)[OF this] have "less_eq (ls ! i) (rs ! i)" "less (bs ! i) (rs ! i)" by auto }
  hence "lex_less_bounded' (bs @ e3 # t3) (ls @ e1 # t1) (rs @ e2 # t2)" using 1 by auto
  then show ?case using 1 1(7)[of 0] by auto
qed (insert e bnd,auto)

end

context non_inf_order begin
abbreviation less where "less x y \<equiv> (y,x) \<in> S"
abbreviation less_eq where "less_eq x y \<equiv> (y,x) \<in> NS"
interpretation ord_syntax less_eq less. (* TODO: remove this line *)

abbreviation less_bnd where
  "less_bnd b \<equiv> {(x,y). (less x y) \<and> (less b y)}"
abbreviation leq_bnd where
  "leq_bnd b \<equiv> {(x,y). (less_eq x y) \<and> (less b y)}"
abbreviation "lex_leq  l1 l2 \<equiv> lex_less_eq l1 l2"

lemma lex_less_empty[simp]:
  "lex_less_bounded' x y [] = False"
  "lex_less_bounded' x [] y = False"
  "lex_less_bounded' [] x y = False"
  by(cases x;cases y;force)+

lemma l1_helper: "x = y \<Longrightarrow> lex_leq x y"
  by(induct x y rule:lex_less_eq.induct,insert refl_NS[unfolded refl_on_def],auto)
lemma l1: "lex_leq x x" using l1_helper by auto

lemma l2: shows "lex_leq xs ys \<Longrightarrow> lex_leq ys zs \<Longrightarrow> lex_leq xs zs"
proof(induct xs zs arbitrary:ys rule:lex_less_eq.induct)
  case (1 l ls r rs yys)
  show ?case proof(cases yys)
    case (Cons y ys)
    from 1(1)[of ys] 1(2-)[unfolded this, simplified]
    show "lex_leq (l # ls) (r # rs)" using
         trans_NS_point   [simplified,of r y l]
         trans_S_point    [simplified,of r y l]
         compat_S_NS_point[simplified,of r y l]
         compat_NS_S_point[simplified,of r y l] by auto
  qed (insert 1, auto)
next case 3 thus ?case by (cases ys) auto
next case 4 thus ?case by (cases ys) auto
qed auto

lemma l3: shows "lex_less_bounded' bounds x y \<Longrightarrow> lex_less_bounded' bounds y z \<Longrightarrow> lex_less_bounded' bounds x z"
proof(induct x z bounds arbitrary:y rule:lex_less_bounded'.induct)
  case (1 l ls r rs b bs)
  thus ?case proof(cases y)
    case (Cons a list)
    with trans_NS_point   [simplified,of r a l]
         trans_S_point    [simplified,of r a l]
         compat_S_NS_point[simplified,of r a l]
         compat_NS_S_point[simplified,of r a l]
         compat_S_NS_point[simplified,of r a b]
    show  "(\<And>y. lex_less_bounded' bs ls y \<Longrightarrow> lex_less_bounded' bs y rs \<Longrightarrow> lex_less_bounded' bs ls rs) \<Longrightarrow>
              lex_less_bounded' (b # bs) (l # ls) y \<Longrightarrow>
              lex_less_bounded' (b # bs) y (r # rs) \<Longrightarrow> y = a # list \<Longrightarrow> lex_less_bounded' (b # bs) (l # ls) (r # rs)"
      by auto
  qed simp
qed auto

lemma l_h: shows "lex_less_bounded' z x y \<Longrightarrow> lex_leq x y"
  by(induct z x y rule:lex_less_bounded'.induct, auto)
  
lemma l4: shows "lex_leq x y \<Longrightarrow> lex_less_bounded' bounds y z \<Longrightarrow> lex_less_bounded' bounds x z"
proof(induct x z bounds arbitrary:y rule:lex_less_bounded'.induct)
  case (1 l ls r rs b bs)
  thus ?case proof(cases y)
    case (Cons a list)
      with trans_NS_point   [simplified,of r a l]
           trans_S_point    [simplified,of r a l]
           compat_S_NS_point[simplified,of r a l]
           compat_NS_S_point[simplified,of r a l] 1
    show ?thesis by auto
  qed simp
next
  case "2_1" thus ?case by(cases y;auto)
qed auto

lemma l5: shows "lex_less_bounded' bounds x y \<Longrightarrow> lex_leq y z \<Longrightarrow> lex_less_bounded' bounds x z" 
proof(induct x z bounds arbitrary:y rule:lex_less_bounded'.induct)
  case (1 l ls r rs b bs)
  thus ?case proof(cases y)
    case (Cons a list)
     with trans_NS_point   [simplified,of r a l]
          trans_S_point    [simplified,of r a l]
          compat_S_NS_point[simplified,of r a l]
          compat_NS_S_point[simplified,of r a l]
          trans_S_point    [simplified,of r a b]
          compat_NS_S_point[simplified,of r a b]  1
    show ?thesis by auto
  qed simp
  next case "2_2" thus ?case by(cases y;auto)
qed auto


lemma less_bnd_is_bound_on :
  "(less_bnd bnd)\<inverse> = bound_on bnd {(x, y). less y x}"
  unfolding bound_on_def
  by (auto simp:bound_on_def)
  
fun restructure where
  "restructure ((a,as),(b,bs)) = (a#as,b#bs)"

lemma restructure_prod:
  "restructure = map_prod (\<lambda> x.(fst x#snd x)) (\<lambda> x.(fst x#snd x))" by auto
  
abbreviation lex2 where
  "lex2 b bs \<equiv> lex_two (less_bnd b) (leq_bnd b) {(x,y). lex_less_bounded' bs x y}"

fun strict_lex_rel where
  "strict_lex_rel b bs = restructure`lex2 b bs"

lemma set_syntax[simp]:
  "greater_eq_set_syntax = NS"
  "greater_set_syntax = S" by auto

lemma lex_less_is_lex:
  "lex_less_bounded' (b#bs) (h1#ls1) (h2#ls2) = 
   (((h1#ls1),(h2#ls2)) \<in> strict_lex_rel b bs)"
  (is "?d = _")
proof -
  have "?d = (((h1,ls1),(h2,ls2)) \<in> lex2 b bs)" by auto
  also have "\<And> A.((h1#ls1),(h2#ls2)) \<in> restructure`A \<longleftrightarrow> ((h1,ls1),(h2,ls2))\<in> A" by force
  finally show ?thesis using strict_lex_rel.simps by blast
qed

lemma restructure_inj:
  "inj (\<lambda> x.((fst x::'x)#snd x))"
proof fix x y
  show "(fst x::'x) # snd x = fst y # snd y \<Longrightarrow> x = y" by(cases x;cases y;auto)
qed

lemma lex_two_converse [simp]:
  "(lex_two a b c)\<inverse> = lex_two (a\<inverse>) (b\<inverse>) (c\<inverse>)"
  by auto
  
lemma restructure_rev_distr [simp]:
  "(restructure ` r)\<inverse> = restructure ` r\<inverse>"
  by (auto simp:restructure_prod)
 
lemma restructure_preserves_SN:
  assumes "SN r"
  shows "SN (restructure ` r)"
  using assms unfolding SN_iff_wf
  using wf_map_prod_image[OF _ restructure_inj,folded restructure_prod]
  by auto
    
lemma lex_wf: "wf {(x,y). lex_less_bounded' bounds x y}"
  (is "wf {(x,y). ?expr x y}")
proof -
  have "SN ({(x,y). ?expr x y}\<inverse>)"
  proof(induct bounds) case Nil thus ?case by auto
  next
    case (Cons a int_bounds)
    {fix b
      have "case b of (x,y) \<Rightarrow> lex_less_bounded' (a # int_bounds) x y =
          (case b of (h1#ls1,h2#ls2) \<Rightarrow> lex_less_bounded' (a # int_bounds) (h1#ls1) (h2#ls2) | _ \<Rightarrow> False)"
        by (auto split:prod.split list.split simp del:lex_less_bounded'.simps)
      hence "(b\<in>{b. case b of (x,y) \<Rightarrow> lex_less_bounded' (a # int_bounds) x y}) = (b\<in>strict_lex_rel a int_bounds)"
        unfolding lex_less_is_lex using lex_less_empty lex_less_is_lex
        by (cases "fst b";cases "snd b";auto simp del:strict_lex_rel.simps) auto
    }
    hence r:"{b. case b of (x,y) \<Rightarrow> lex_less_bounded' (a # int_bounds) x y} = strict_lex_rel a int_bounds"
      by simp
    show ?case unfolding r strict_lex_rel.simps restructure_rev_distr lex_two_converse
      proof (rule restructure_preserves_SN,rule lex_two,goal_cases)
        case 1 thus ?case using compat_NS_S_point by auto
      next
        case 2
        then show ?case unfolding less_bnd_is_bound_on
              using bound_on_SN by auto
      next
        case 3 thus ?case using Cons by auto
      qed
  qed
  thus ?thesis using SN_iff_wf[of "{(x,y).?expr y x}"] by auto
qed

lemma lex_wf_order: "class.wf_order lex_less_eq (lex_less_bounded' bounds)"
 proof
  fix P a
  assume a1:"(\<And>x. (\<And>y. lex_less_bounded' bounds y x \<Longrightarrow> P y) \<Longrightarrow> P x)"
  thus "P a"
 by (rule lex_wf[unfolded wf_def,rule_format]) blast
qed (fact l1 l2 l3 l4 l5)+

lemma wf_grouped: "wf {(x,y). lex_less_bounded' bounds (lex_group x) (lex_group y)}"
  (is "wf ?r")
proof -
  have r:"?r = inv_image {(x,y). lex_less_bounded' bounds x y} lex_group" by auto
  show ?thesis unfolding r by(fact wf_inv_image[OF lex_wf])
qed

lemma wf_order_grouped:"class.wf_order (\<lambda>x y. lex_leq (lex_group x) (lex_group y)) (\<lambda>x y. lex_less_bounded' bounds (lex_group x) (lex_group y))"
proof fix P a
  assume a1:"\<And>x. (\<And>y. lex_less_bounded' bounds (lex_group y) (lex_group x) \<Longrightarrow> P y) \<Longrightarrow> P x"
  thus "P a" using wf_grouped[unfolded wf_def,rule_format, of bounds lex_group] by auto
qed (fact l1 l2 l3 l4 l5)+

end

section {* Logic for lexicographic orders *}

datatype 't Lex = LList (lex_list:"'t list")
    | LEmbed (lex_elem:'t)
    | LBool
datatype ('f,'t) LexF = Lex_LessEq "'t list" | Lex_Less "'t list" | Lex_Group "'t list" | Lex_Embed (lex_embed:'f)
datatype 'd LexV = Lex_GroupV (lex_group:"'d list") | Embed_In_Lex (lex_val:'d) | Lex_Bool bool

instantiation Lex :: ("show") "show"
begin

 (* TODO: replace by 
derive "show" Lex
derive "show" LexF
 when that works *)
definition
 "shows_prec (p :: nat) v = (case v of
    (LBool) \<Rightarrow> shows  ''LBool '' |
    (LList e) \<Rightarrow> shows  ''LList '' o shows e |
    (LEmbed e) \<Rightarrow> shows ''LBool '' o shows e)"
 
lemma shows_prec_Lex_append [show_law_simps]:
  "shows_prec p (f ::'a Lex) (r @ s) = shows_prec p f r @ s"
  by (cases f; simp add: shows_prec_Lex_def show_law_simps)

definition "shows_list (xss :: 'a Lex list) = showsp_list shows_prec 0 xss"

instance by standard (simp_all add: shows_list_Lex_def show_law_simps)
end

instantiation LexF :: ("show", "show") "show"
begin
  
definition
 "shows_prec (p :: nat) v = (case v of
    (Lex_LessEq   e) \<Rightarrow> shows ''Lex_LessEq '' o shows e |
    (Lex_Less   e) \<Rightarrow> shows ''Lex_Less '' o shows e |
    (Lex_Group   e) \<Rightarrow> shows ''Lex_Group '' o shows e |
    (Lex_Embed   e) \<Rightarrow> shows ''Lex_Embed '' o shows e)"

lemma shows_prec_LexF_append [show_law_simps]:
  "shows_prec p (f ::('a,'b) LexF) (r @ s) = shows_prec p f r @ s"
  by (cases f; simp add: shows_prec_LexF_def show_law_simps)
 
definition "shows_list (xss :: ('a,'b) LexF list) = showsp_list shows_prec 0 xss"

instance by standard (simp_all add: shows_list_LexF_def show_law_simps)
end


locale pre_checker_lex =
  base: pre_logic_checker \<Lambda> check_logic sa +
  ord_syntax less_eq less
  for \<Lambda> :: "('f::show,'t::show,'d) logic"
  and check_logic :: "('h::{default,show},'f,'v::show,'t) logic_checker"
  and sa
  and less_eq_formula less_formula :: "('f,'v,'t) exp \<Rightarrow> ('f,'v,'t) exp \<Rightarrow> ('f,'v,'t) exp formula"
  and less_eq less :: "'d \<Rightarrow> 'd \<Rightarrow> bool"
  and bounds :: "'d list"
begin

notation less_eq_formula (infix "\<sqsubseteq>\<^sub>f" 50)
notation less_formula (infix "\<sqsubset>\<^sub>f" 50)

abbreviation "lex_leq \<equiv> lex_less_eq"
abbreviation "lex_less \<equiv> lex_less_bounded' bounds"
abbreviation "zpb l1 l2 \<equiv> zip l1 (zip l2 bounds)"

fun lex_tobool where
"lex_tobool (Embed_In_Lex v) = base.to_bool v" |
"lex_tobool (Lex_Bool v) = v"

fun lex_type_values :: "'t Lex \<Rightarrow> 'd LexV set" where
"lex_type_values (LList tp) = Lex_GroupV ` listset (map base.Values_of_type tp)" |
"lex_type_values (LEmbed v) = Embed_In_Lex ` base.Values_of_type v" |
"lex_type_values (LBool) = {Lex_Bool True,Lex_Bool False}"

fun lex_types :: "('f,'t) LexF \<Rightarrow> 't Lex list \<times> 't Lex" where
"lex_types (Lex_LessEq tps) = ([LList tps,LList tps],LBool)" |
"lex_types (Lex_Less tps)   = ([LList tps,LList tps],LBool)" |
"lex_types (Lex_Group tps)  = (map LEmbed tps,LList tps)" |
"lex_types (Lex_Embed tp)  = (let (args_LIA,tp_LIA) = base.type_of_fun tp in (map LEmbed (args_LIA),LEmbed tp_LIA))"

fun lex_interpret :: "('f, 't) LexF \<Rightarrow> 'd LexV list \<Rightarrow> 'd LexV" where
  "lex_interpret (Lex_LessEq _) [v1,v2] = Lex_Bool (lex_leq (lex_group v1) (lex_group v2))" |
  "lex_interpret (Lex_Less  _)  [v1,v2] = Lex_Bool (lex_less (lex_group v1) (lex_group v2))" |
  "lex_interpret (Lex_Group _)   vs = Lex_GroupV (map lex_val vs)" |
  "lex_interpret (Lex_Embed f)   vs = Embed_In_Lex (base.interpret f (map lex_val vs))"

definition lex_logic :: "(('f,'t) LexF, 't Lex, 'd LexV) Formula.logic" where
  "lex_logic =  \<lparr>
  logic.type_of_fun = lex_types,
  Values_of_type = lex_type_values,
  Bool_types = LEmbed ` base.Bool_types \<union> {LBool},
  to_bool = lex_tobool,
  interpret = lex_interpret
\<rparr>"
interpretation prelogic lex_logic.

abbreviation lex_leq_formula :: (* 'e needs to be a free type *)
  "      (('f, 't) LexF, 'e \<times> 't Lex) Term.term
      \<Rightarrow> (('f, 't) LexF, 'e \<times> 't Lex) Term.term
      \<Rightarrow> (('f, 't) LexF, 'e \<times> 't Lex) Term.term formula"
  where "lex_leq_formula t1 t2 \<equiv> Atom (Fun (Lex_LessEq (lex_list (THE ty. t1 :\<^sub>f ty))) [t1,t2])"
abbreviation lex_less_formula
  where "lex_less_formula t1 t2 \<equiv> Atom (Fun (Lex_Less  (lex_list (THE ty. t1 :\<^sub>f ty))) [t1,t2])"

fun lex_embed_rec :: "('f, 'v \<times> 't) Term.term \<Rightarrow> (('f, 'c) LexF, 'v \<times> 't Lex) Term.term" where
"lex_embed_rec (Fun v lst) = Fun (Lex_Embed v) (map lex_embed_rec lst)" |
"lex_embed_rec (Var (h,t)) = Var (h,LEmbed t)"

lemma lex_embed_rec_is_Fun[simp]:
 "is_Fun (lex_embed_rec f) = is_Fun f" "is_Var (lex_embed_rec f) = is_Var f" by(cases f;auto)+

lemma lex_embed_rec_helper[simp]:
  "(\<And>x2a. x2a \<in> set x2 \<Longrightarrow> f1 (f3 x2a) = map_option LEmbed (f2 x2a)) \<Longrightarrow>
   Option_Monad.mapM f1 (map f3 x2) \<bind> (\<lambda>tys. guard (tys = map LEmbed x1) \<bind> (\<lambda>_. Some (LEmbed x2a))) =
              Option_Monad.mapM f2 x2 \<bind> (\<lambda>x. guard (x = x1) \<bind> (\<lambda>x. Some (LEmbed x2a)))"
  proof(induct x2 arbitrary:x1)
    case Nil thus ?case by (auto simp:eq_commute[of "[]"])
  next
    case (Cons a x2)
    have ind:"(\<And>x2a. x2a \<in> set x2 \<Longrightarrow> f1 (f3 x2a) = map_option LEmbed (f2 x2a))"
         and a:"f1 (f3 a) = map_option LEmbed (f2 a)"
      using Cons(2) by auto
    note ind=Cons(1)[OF ind,simplified]
    show ?case proof(cases x1)
      case Nil show ?thesis unfolding Nil by auto
    next
      case (Cons a2 list){ fix P g fix g2::"'b list \<Rightarrow> 'free_tp option"
        assume "(Option_Monad.mapM f1 (map f3 x2) \<bind> g) = (Option_Monad.mapM f2 x2 \<bind> g2)"
        hence "(Option_Monad.mapM f1 (map f3 x2) \<bind> (\<lambda>ya. guard P \<bind> (\<lambda> _. g ya))) =
                        (Option_Monad.mapM f2 x2 \<bind> (\<lambda>ya. guard P \<bind> (\<lambda> _. g2 ya)))"
          by (cases P;auto)}
      from this[OF ind[of list]]
      show ?thesis unfolding list.map mapM.simps Option.bind_assoc Cons a
         by (cases "f2 a",auto simp:guard_and_to_bind)
    qed
  qed

fun lex_unembed_rec :: "(('f, 'c) LexF, 'v \<times> 't Lex) Term.term \<Rightarrow> ('f, 'v \<times> 't) Term.term" where
"lex_unembed_rec (Fun (Lex_Embed v) lst) = Fun v (map lex_unembed_rec lst)" |
"lex_unembed_rec (Var (h,LEmbed t)) = Var (h, t)"

lemma lex_embed_unembed[simp]:
 "lex_unembed_rec (lex_embed_rec f) = f" by(induct f;auto simp:map_idI)

definition check_logic_lex where
  "check_logic_lex h \<phi> = error (shows ''lexicographic logic checker not implemented'')"

end

locale checker_lex = pre_checker_lex \<Lambda> check_logic sa less_eq_formula less_formula less_eq less bounds + 
  base: logic_checker \<Lambda> check_logic sa
  for \<Lambda> :: "('f::show,'t::show,'d) logic"
  and check_logic :: "('h::{default,show},'f,'v::show,'t) logic_checker"
  and sa
  and less_eq_formula less_formula :: "('f,'v,'t) exp \<Rightarrow> ('f,'v,'t) exp \<Rightarrow> ('f,'v,'t) exp formula"
  and less_eq less :: "'d \<Rightarrow> 'd \<Rightarrow> bool"
  and bound_exps :: "('f,'w,'t) exp list"
  and bounds :: "'d list" +
  assumes ni: "non_inf_order {(x,y). less y x} {(x,y). less_eq y x}"
     and bounds: "\<And> \<alpha>. bounds = map (\<lambda> v. base.eval v \<alpha>) bound_exps"
begin
interpretation prelogic lex_logic.
abbreviation bound_types :: "'t list" where
  "bound_types \<equiv> map (\<lambda>e. THE ty. base.has_type e ty) bound_exps"
lemma lex_logic_fold:
  "lex_type_values = Values_of_type"
  "lex_types = type_of_fun"
  "lex_tobool = to_bool"
  "lex_interpret = interpret" by (auto simp:lex_logic_def)

lemmas lex_logic_simps[unfolded lex_logic_fold, simp] =
  lex_type_values.simps
  lex_types.simps
  lex_tobool.simps
  lex_interpret.simps

lemma mem_bool_simps[simp]:
  "ty \<in> Bool_types \<longleftrightarrow> ty \<in> LEmbed ` base.Bool_types \<union> {LBool}" by (auto simp:lex_logic_def)

lemma consistent_listset[simp]:"listset (map base.Values_of_type list) \<noteq> {}"
    by(induct list,insert base.consistent,auto simp:set_Cons_def)


interpretation nio:non_inf_order "{(x,y). less y x}" "{(x,y). less_eq y x}" using ni(1).

lemma lex_logic:"logic_checker lex_logic check_logic_lex"
proof(unfold_locales)
  fix ds f
  assume len: "length ds = length (param_types f)"
     and i: "i < length ds \<Longrightarrow> ds ! i \<in> Values_of_type (param_types f ! i)" for i

  show "interpret f ds \<in> local.Values_of_type (return_type f)"
  proof (cases f)
    case (Lex_LessEq x1)
    with len i base.well_typed show ?thesis by (cases ds rule:list_4_cases; auto)
  next
   case (Lex_Less x2)
   with len i base.well_typed show ?thesis by (cases ds rule: list_4_cases; auto)
  next
    case [simp]: (Lex_Group x3)
    from len[simplified] i[simplified] have "map lex_val ds \<in> listset (map base.Values_of_type x3)"
    proof (induct x3 arbitrary: ds)
     case Nil
     then show ?case by (auto simp: image_def)
    next
     case IH: (Cons a x3)
     show ?case
     proof (cases ds)
       case Nil with IH show ?thesis by auto
     next
       case [simp]: (Cons d ds')
       show ?thesis using IH(3)[of 0]
        apply (simp add: image_def set_Cons_def)
        apply (intro conjI IH(1), force) using  IH(2) apply simp
        using IH(3)[simplified] apply force done
     qed
    qed
    then show ?thesis by auto
  next
    case [simp]: (Lex_Embed f2)
    show ?thesis
    proof(cases "base.type_of_fun f2")
      case [simp]: (Pair args ret)
      note len[simplified,simp]
      note [intro!] = base.well_typed[of "map lex_val ds" "f2", simplified]
      { fix i assume "i < length args"
        from i[simplified, OF this, unfolded nth_map[OF this]]
        have "lex_val (ds ! i) \<in> base.Values_of_type (args ! i)" by auto
      }
      then show ?thesis by auto
     qed
  qed
next
  fix ty
  show "local.Values_of_type ty \<noteq> {}" by (induct ty, insert base.consistent, auto)
next
qed (auto simp: check_logic_lex_def)

sublocale logic_checker lex_logic check_logic_lex by (fact lex_logic)

interpretation lex_order: wf_order "\<lambda>x y. lex_leq (lex_group x) (lex_group y)" "\<lambda>x y. lex_less (lex_group x) (lex_group y)"
  using non_inf_order.wf_order_grouped[OF ni(1)] by auto


lemma mapM_map_eq_Some[simp]: (* TODO: move *)
  shows "Option_Monad.mapM g xs = Some (map f ys) \<longleftrightarrow> map (Some o f) ys = map g xs"
    proof(induct xs arbitrary:ys)
      case Cons thus ?case by(cases ys; auto)
    qed auto

lemma eval_rename[simp]: (* TODO: move *)
  "base.eval z (\<lambda>(x, y). \<alpha> (f x, y)) = base.eval (rename_vars f z) \<alpha>"
proof(induct z)
  case (Fun x1a x2)
  then show ?case by(auto simp:o_def intro:arg_cong[of _ _ "base.interpret x1a"])
qed auto


end

end
