(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016, 2017)
License: LGPL (see file COPYING.LESSER)
*)

theory Gtcap_Impl
  imports
    "../Rewriting/Tcap_Impl"
    Gtcap
    "$AFP/Transitive-Closure/Transitive_Closure_List_Impl"
    "../Rewriting/Term_Impl"
begin

definition "gt1 = map (\<lambda>r. case r of (s, t) \<Rightarrow> (root s, root t))"

lemma gt1_iff_GT1 [simp]:
  "set (gt1 rs) = GT1 (set rs)"
by (auto simp: GT1_def gt1_def)

definition "mk_gt_fun rs =
  (let in_trancl = memo_list_trancl (gt1 rs) in
  (\<lambda>f g. g \<in> set (in_trancl f)))"

lemma mk_gt_fun_iff [simp]:
  "mk_gt_fun rs f g \<longleftrightarrow> (f, g) \<in> (GT1 (set rs))\<^sup>+"
by (auto simp: mk_gt_fun_def memo_list_trancl)

lemma GT_set_iff:
  "(s, t) \<in> GT (set rs) \<longleftrightarrow>
  is_Var s \<or> is_Var t \<or> mk_gt_fun rs (root s) (root t) \<or> mk_gt_fun rs (root s) None
  \<or> mk_gt_fun rs None (root t) \<or> mk_gt_fun rs None None"
by (auto simp: GT_def)

definition gt_term
where
  "gt_term nlv ne gt_fun rm s t \<longleftrightarrow> ne \<and>
    (let root1 = root s; root2 = root t in
    (is_Var s \<or> is_Var t \<or> gt_fun None None \<or> gt_fun root1 root2 \<or> gt_fun root1 None \<or> gt_fun None root2) \<and>
    (if nlv then
      (case root1 of
        Some fn \<Rightarrow> \<exists>r\<in>set (rm fn). Ground_Context.match (tcapRM nlv rm s) (fst r)
      | None \<Rightarrow> True)
    else True))"

lemma root_None_gc_match:
  "root s = None \<Longrightarrow> Ground_Context.match (tcap R s) t"
by (cases s) (auto simp: Ground_Context.match_def)

lemma ne_pair_list:
  "xs \<noteq> [] \<Longrightarrow> \<exists>x y. (x, y) \<in> set xs"
by (cases xs) auto

lemma varcond_root:
  "\<forall>lr \<in> R. is_Fun (fst lr) \<Longrightarrow> (x, y) \<in> R \<Longrightarrow> \<exists>f n. root x = Some (f, n)"
by (cases x) auto

lemma var_lhs_gc_match:
  "(l, r) \<in> R \<Longrightarrow> is_Var l \<Longrightarrow> Ground_Context.match (tcap R s) l"
apply (cases l) apply (auto simp: Ground_Context.match_def)
by (meson tcap_instance_equiv_class)

lemma gt_term_iff:
  assumes nlv: "nlv \<longleftrightarrow> (\<forall>lr \<in> set rs. is_Fun (fst lr))"
    and rm: "\<And>fn. set (rm fn) = {(l, r). (l, r) \<in> set rs \<and> root l = Some fn}"
  shows "gt_term nlv (rs \<noteq> []) (mk_gt_fun rs) rm s t \<longleftrightarrow>
    (s, t) \<in> GT (set rs) \<inter>
    {(s, t :: ('f, 'v) term). \<exists>(l, r) \<in> set rs. Ground_Context.match (tcap (set rs) s) l}"
proof (cases nlv)
  case True
  then have *: "\<forall>lr \<in> set rs. is_Fun (fst lr)" using nlv by auto
  show ?thesis
    using True and varcond_root [OF *]
    apply (simp add: gt_term_def Let_def tcapRM [OF assms, simplified True] GT_set_iff rm split: option.splits)
    apply (cases s; cases t; simp)
       apply (force simp: Ground_Context.match_def dest: ne_pair_list)
      apply (force simp: Ground_Context.match_def dest: ne_pair_list)
     apply (auto simp: Let_def Ground_Context.match_def)
    using nlv apply fastforce
            apply (metis fst_conv)
    using nlv apply fastforce
    using nlv apply fastforce
    using nlv apply fastforce
    using nlv apply fastforce
       apply (metis fst_conv)
      apply (metis fst_conv)
     apply (metis fst_conv)
    apply (metis fst_conv)
    done
next
  case False
  with nlv obtain l and r where "(l, r) \<in> set rs" and "is_Var l" by auto
  then have "Ground_Context.match (tcap (set rs) s) l"
    using var_lhs_gc_match by auto
  then show ?thesis
    using False and \<open>(l, r) \<in> set rs\<close>
    by (auto simp: gt_term_def Let_def GT_def)
qed

fun rd_impl
where
  "rd_impl gt (Fun f ss, Fun g ts) =
    (if f = g \<and> length ss = length ts \<and> \<not> gt (Fun f ss) (Fun g ts) then concat (map (rd_impl gt) (zip ss ts))
    else [(Fun f ss, Fun g ts)])"
| "rd_impl _ (s, t) = [(s, t)]"

lemma rd_impl:
  fixes R :: "('f, 'v) trs"
  assumes gt: "\<And>s t. gt s t \<longleftrightarrow>
    (s, t) \<in> GT R \<inter> {(s, t :: ('f, 'v) term). \<exists>(l, r) \<in> R. Ground_Context.match (tcap R s) l}"
    (is "\<And>s t. gt s t \<longleftrightarrow> (s, t) \<in> ?S")
    and "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
  shows "\<forall>(u, v) \<in> set (rd_impl gt (s, t)). (u \<cdot> \<sigma>, v \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof -
  interpret reachability_decomposition R "?S"
    by (intro reachability_decomposition_Int rd_root rd_tcap)
  have "\<And>s t. rd_impl gt (s, t) = reach_decomp s t"
  proof -
    fix s t
    show "?thesis s t"
    unfolding reach_decomp_reach_decomp'_conv
    apply (induct ("(s, t)") rule: reach_decomp'.induct)
    apply (case_tac "(Fun f ss, Fun g ts) \<in> ?S")
    apply (auto simp only: reach_decomp'.simps rd_impl.simps gt [symmetric])
    apply (auto intro!: arg_cong [of _ _ concat])
    done
  qed
  then show ?thesis using assms by (auto dest: reach_decomp_sound)
qed

definition nonreach
  where
    "nonreach gt s t =
      (case (s, t) of
        (Fun f ss, Fun g ts) \<Rightarrow> (f, length ss) \<noteq> (g, length ts) \<and> \<not> gt s t
      | _ \<Rightarrow> False)"

lemma nonreach:
  fixes R :: "('f, 'v) trs"
  assumes gt: "\<And>s t. gt s t \<longleftrightarrow>
    (s, t) \<in> GT R \<inter> {(s, t :: ('f, 'v) term). \<exists>(l, r) \<in> R. Ground_Context.match (tcap R s) l}"
    (is "\<And>s t. gt s t \<longleftrightarrow> (s, t) \<in> ?S")
  assumes nr: "nonreach gt s t"
    and *: "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*" (is "(?s, ?t) \<in> _")
  shows False
proof -
  obtain f ss and g ts where [simp]: "s = Fun f ss" "t = Fun g ts"
    and ne: "(f, length ss) \<noteq> (g, length ts)" and "\<not> gt s t"
    using nr by (auto simp: nonreach_def split: term.splits)
  have "(?s, ?t) \<in> GE R" using rsteps_GE [OF *] .
  then have "(?s, ?t) \<in> GT R" using ne by auto
  then have GT: "(s, t) \<in> GT R" by (rule GT_substs)

  obtain n where "rrsteps_n R (Suc n) ?s ?t"
    using rsteps_imp_rrsteps_n [OF *] and rrsteps_n_0 [of R ?s ?t] and ne
    by auto (metis old.nat.exhaust)+
  from rrsteps_n_match_tcap [OF this] and \<open>\<not> gt s t\<close> [unfolded gt]
  show False
    using GT and tcap_substs [of s \<sigma> t \<tau> R]
    by (auto simp del: tcap.simps simp: gt)
qed

definition "nonreachable_gtcapRM nlv ne gt_fun rm s t =
  (let gt = gt_term nlv ne gt_fun rm in
  (\<exists>(u, v) \<in> set (rd_impl gt (s, t)). nonreach gt u v))"

lemma nonreach_gtcap:
  fixes R :: "('f, 'v) rules" and s t :: "('f, 'v) term"
  assumes nlv: "nlv \<longleftrightarrow> (\<forall>lr\<in>set R. is_Fun (fst lr))" (is "_ = ?nlv")
    and rm: "\<And>fn. set (rm fn) = {(l, r) \<in> set R. root l = Some fn}"
  assumes nr: "nonreachable_gtcapRM nlv (R \<noteq> []) (mk_gt_fun R) rm s t"
    and *: "(s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*"
  shows False
proof -
  let ?gt = "gt_term nlv (R \<noteq> []) (mk_gt_fun R) rm"
  let ?rd = "rd_impl ?gt (s, t)"
  from rd_impl [OF gt_term_iff, OF nlv rm *]
  have "\<forall>(u, v) \<in> set ?rd. (u \<cdot> \<sigma>, v \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*" by (simp add: nlv)
  obtain u and v where "(u, v) \<in> set ?rd" and nr': "nonreach ?gt u v"
    using nr by (auto simp: nonreachable_gtcapRM_def Let_def)
  moreover
  have "(u \<cdot> \<sigma>, v \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*"
    using rd_impl [OF gt_term_iff, OF nlv rm *] and \<open>(u, v) \<in> set ?rd\<close> by (auto simp: nlv)
  ultimately
  show False
    using nonreach [OF gt_term_iff, OF nlv rm nr', of \<sigma> \<tau>]
    by simp
qed

end
